# XV6网络模块设计文档

我们最初的想法是实现基本网络ping工具，我们实现了虚拟网络设备、链路层、网络层、传输层的收发函数，实现了网卡初始化、ARP请求和响应、ICMP消息的封装与解析等等功能模块。

但很遗憾最终没有完全实现网络功能，成功通过了编译，但经过了很多轮测试，仍然无法ping通，最终只能是作为未完成的分支发布，没有合并到主分支。具体分支请见：[网络分支](https://gitlab.eduxiji.net/T202410336994286/project2608132-273224/-/tree/lbzdev2?ref_type=heads)

以下是XV6网络模块的结构树（逻辑结构而非文件结构）:

```
xv6-network/
├── kernel/
│   ├── netdev/ 
│   │   ├── netdev.h      # 网络设备接口定义
│   │   ├── netdev.c      # 网络设备通用操作实现
│   │   └── virtio_net.c  # virtio网卡驱动实现
│   │
│   ├── ethernet/
│   │   ├── ethernet.h    # 以太网协议头文件
│   │   └── ethernet.c    # 以太网协议实现
│   │
│   ├── arp/
│   │   ├── arp.h        # ARP协议头文件
│   │   └── arp.c        # ARP协议实现
│   │
│   ├── ip/
│   │   ├── ip.h         # IP协议头文件
│   │   └── ip.c         # IP协议实现
│   │
│   ├── icmp/
│   │   ├── icmp.h       # ICMP协议头文件
│   │   └── icmp.c       # ICMP协议实现
│   │
│   ├── tcp/
│   │   ├── tcp.h        # TCP协议头文件
│   │   └── tcp.c        # TCP协议实现
│   │
│   ├── udp/
│   │   ├── udp.h        # UDP协议头文件  
│   │   └── udp.c        # UDP协议实现
│   │
│   ├── socket/
│   │   ├── socket.h     # Socket接口定义
│   │   └── socket.c     # Socket接口实现
│   │
│   └── net/
│       ├── net.h        # 网络公共头文件
│       └── netdef.h     # 网络相关数据结构定义
│
└── user/
    ├── net.h           # 用户态网络头文件
    ├── ping.c          # ping工具实现
    └── netcat.c        # netcat工具实现
```

主要模块说明:

1. netdev/ - 网络设备驱动层
2. ethernet/ - 以太网协议层
3. arp/ - ARP协议实现
4. ip/ - IP协议实现
5. icmp/ - ICMP协议实现 
6. tcp/ - TCP协议实现
7. udp/ - UDP协议实现
8. socket/ - Socket接口实现
9. net/ - 通用网络定义
10. user/ - 用户态网络工具和测试程序

## 1. 网络设备层(Network Device Layer)

### 功能
提供网络设备的抽象接口,包括设备初始化、数据包发送和接收等基础功能。是整个网络栈的最底层。

### 结构
- 定义了统一的网络设备结构体(netdev)
- 提供设备操作的函数指针接口
- 支持基本的设备锁机制
- 维护设备的MAC地址和IP地址

### 代码简介
核心数据结构:
```c
struct netdev {
    unsigned char mac[6];     // MAC地址
    unsigned int ip;          // IP地址 
    struct spinlock lock;     // 设备锁
    
    // 设备操作函数指针
    int (*init)(struct netdev *dev);
    int (*transmit)(struct netdev *dev, void *data, unsigned short len);
    int (*receive)(struct netdev *dev, void *data, unsigned short len);
    
    void *priv;              // 设备私有数据
};
```

核心方法:
- netdev_init(): 初始化网络设备
- netdev_transmit(): 发送数据包
- netdev_receive(): 接收数据包
- netdev_loopback(): 本地回环处理

## 2. 以太网层(Ethernet Layer)

### 功能
处理以太网帧的封装和解析,实现MAC层的基本功能。

### 结构  
- 定义以太网帧格式
- 处理以太网协议类型
- 提供帧的发送和接收接口

### 代码简介
核心数据结构:
```c
struct eth_hdr {
    unsigned char dst[ETH_ADDR_LEN]; // 目标MAC
    unsigned char src[ETH_ADDR_LEN]; // 源MAC  
    unsigned short type;            // 协议类型
};

struct eth_frame {
    struct eth_hdr hdr;
    unsigned char payload[ETH_FRAME_LEN - ETH_HDR_LEN];
};
```

核心方法:
- eth_init(): 以太网初始化
- eth_send(): 发送以太网帧
- eth_receive(): 接收以太网帧
- eth_input(): 以太网帧输入处理

## 3. ARP协议模块(Address Resolution Protocol)

### 功能
实现IP地址到MAC地址的解析和缓存管理。

### 结构
- ARP协议报文的处理
- ARP缓存表的维护
- ARP请求和响应机制

### 代码简介
核心数据结构:
```c
struct arp_hdr {
    unsigned short hwtype;           // 硬件类型
    unsigned short protype;          // 协议类型
    unsigned char hwlen;            // 硬件地址长度
    unsigned char prolen;           // 协议地址长度
    unsigned short op;              // 操作类型
    unsigned char sha[ETH_ADDR_LEN];// 发送方硬件地址
    unsigned int spa;              // 发送方协议地址
    unsigned char tha[ETH_ADDR_LEN];// 目标硬件地址
    unsigned int tpa;              // 目标协议地址
};

struct arp_entry {
    unsigned int ip;              // IP地址
    unsigned char mac[ETH_ADDR_LEN]; // MAC地址
    unsigned int time;            // 更新时间  
    int valid;                    // 是否有效
};
```

核心方法:
- arp_init(): 初始化ARP模块
- arp_request(): 发送ARP请求
- arp_input(): 处理接收的ARP包
- arp_lookup(): 查找ARP缓存
- arp_update(): 更新ARP缓存

## 4. IP协议模块(Internet Protocol)

### 功能
实现网络层的数据包封装、路由和分片重组。

### 结构
- IP数据包的封装与解析
- IP校验和计算
- IP地址管理
- 协议分发处理

### 代码简介
核心数据结构:
```c
struct ip_hdr {
    unsigned char ver_ihl;    // 版本和头长度
    unsigned char tos;        // 服务类型
    unsigned short len;       // 总长度 
    unsigned short id;        // 标识
    unsigned short frag_off;  // 分片偏移
    unsigned char ttl;        // 生存时间
    unsigned char proto;      // 协议
    unsigned short checksum;  // 校验和
    unsigned int saddr;      // 源IP
    unsigned int daddr;      // 目标IP
};
```

核心方法:
- ip_init(): 初始化IP模块 
- ip_send(): 发送IP数据包
- ip_input(): 处理接收的IP包
- ip_checksum(): IP校验和计算

## 5. ICMP协议模块(Internet Control Message Protocol)

### 功能
实现网络控制消息协议,用于网络诊断和错误报告。

### 结构
- ICMP消息的封装与解析
- 处理Echo请求/应答
- 处理各类ICMP错误消息

### 代码简介
核心数据结构:
```c
struct icmp_hdr {
    unsigned char type;      // 类型
    unsigned char code;      // 代码
    unsigned short checksum; // 校验和
    unsigned short id;       // 标识符
    unsigned short seq;      // 序列号
};
```

核心方法:
- icmp_init(): 初始化ICMP模块
- icmp_input(): 处理接收的ICMP包
- icmp_send_echo(): 发送Echo请求
- icmp_checksum(): ICMP校验和计算

## 6. TCP协议模块(Transmission Control Protocol) 

### 功能
实现面向连接的可靠数据传输。

### 结构
- TCP连接管理
- 流量控制
- 拥塞控制
- 数据重传机制
- 定时器管理

### 代码简介
核心数据结构:
```c
struct tcp_hdr {
    unsigned short sport;    // 源端口
    unsigned short dport;    // 目标端口
    unsigned int seq;       // 序列号
    unsigned int ack;       // 确认号
    unsigned char offset;   // 数据偏移
    unsigned char flags;    // 标志位
    unsigned short window;  // 窗口大小
    unsigned short checksum;// 校验和
    unsigned short urgptr; // 紧急指针
};

struct tcp_pcb {
    struct tcp_pcb *next;   // PCB链表
    struct netdev *netdev;  // 网络设备
    
    // 连接标识
    unsigned int local_ip;
    unsigned short local_port;
    unsigned int remote_ip;
    unsigned short remote_port;
    
    unsigned char state;    // TCP状态
    
    // 序列号管理
    unsigned int snd_una;  // 未确认序列号
    unsigned int snd_nxt;  // 下一发送序列号
    unsigned int rcv_nxt;  // 期望接收序列号
    
    // 窗口管理
    unsigned short snd_wnd;
    unsigned short rcv_wnd;
    
    // 定时器
    unsigned int timer;
    unsigned int retrans_timer;
    unsigned int persist_timer;
    unsigned int keepalive_timer;
    
    // 重传管理
    unsigned int rto;
    unsigned int retries;
    
    // 拥塞控制
    unsigned char cc_state;
    unsigned int cwnd;
    unsigned int ssthresh;
    
    // 缓冲区
    struct tcp_buf *rcv_buf;
    struct tcp_buf *snd_buf;
};
```

核心方法:
- tcp_init(): 初始化TCP模块
- tcp_connect(): 建立TCP连接
- tcp_input(): 处理接收的TCP段
- tcp_send_segment(): 发送TCP段
- tcp_close(): 关闭TCP连接
- tcp_timer(): TCP定时器处理

## 7. UDP协议模块(User Datagram Protocol)

### 功能
实现无连接的数据报传输服务。

### 结构
- UDP数据报的封装与解析
- UDP端口管理
- 简单的收发接口

### 代码简介
核心数据结构:
```c
struct udp_hdr {
    unsigned short sport;    // 源端口
    unsigned short dport;    // 目标端口  
    unsigned short len;      // UDP长度
    unsigned short checksum; // 校验和
};

struct udp_pcb {
    struct udp_pcb *next;   // PCB链表
    struct netdev *netdev;  // 网络设备
    
    unsigned int local_ip;
    unsigned short local_port;
    unsigned int remote_ip;
    unsigned short remote_port;
    
    void (*recv)(void *arg, struct udp_pcb *pcb, void *data, int len);
    void *recv_arg;
};
```

核心方法:
- udp_init(): 初始化UDP模块
- udp_new(): 创建UDP控制块
- udp_bind(): 绑定本地地址
- udp_send(): 发送UDP数据报
- udp_input(): 处理接收的UDP数据报

## 8. Socket接口层

### 功能
提供BSD socket风格的网络编程接口。

### 结构
- Socket抽象
- 统一的操作接口
- 支持TCP/UDP/RAW socket

### 代码简介
核心数据结构:
```c
struct socket {
    int fd;                // 文件描述符
    int type;             // socket类型
    int state;            // socket状态
    struct tcp_pcb *pcb;  // TCP控制块
    struct spinlock lock; // 锁
};

struct sockaddr {
    uint16 family;
    uint16 port;
    uint32 addr;
    char zero[8];
};
```

核心方法:
- sock_socket(): 创建socket
- sock_bind(): 绑定地址
- sock_listen(): 监听连接
- sock_accept(): 接受连接
- sock_connect(): 建立连接
- sock_close(): 关闭socket
- sock_read(): 读取数据
- sock_write(): 发送数据

## 9. 测试模块
- ping工具:ICMP Echo测试
- netcat工具:TCP/UDP连接测试