#include "types.h"
#include "param.h"
#include "riscv.h"
#include "memlayout.h"
#include "spinlock.h"
#include "defs.h"
#include "proc.h"

int
issig(struct proc *p){
    return p->sigpending & (~(p->sigblocked));
}

int
ign(void){
    return 0;
}

void
coredump(void)
{
  static char *states[] = {
  [UNUSED]    "unused",
  [USED]      "used",
  [SLEEPING]  "sleep ",
  [RUNNABLE]  "runble",
  [RUNNING]   "run   ",
  [ZOMBIE]    "zombie"
  };
  struct proc *p = myproc();
  char *state;

  printf("\n");
    if(p->state != UNUSED){
        if(p->state >= 0 && p->state < NELEM(states) && states[p->state])
      state = states[p->state];
    else
      state = "???";
    printf("%d %s %s", p->pid, state, p->name);
    printf("\n");
    }
}

void
core(struct proc *p, int * done)
{
	coredump();
	release(&proc->lock);
	kill(p->pid);
	acquire(&proc->lock);
	*done = 1;
}

void
term(struct proc *q, int * done)
{
	release(&proc->lock);
	kill(q->pid);
	acquire(&proc->lock);
	*done = 1;
}

void
cont(struct proc *p)
{
		
	if(p->state == SLEEPING)
	{
		p->state = RUNNABLE;
		p->justwoken = 1;
	}
}

void
stop(struct proc *q, int *done)
{	
	q->state = SLEEPING;
	sched();

	*done = 1;
}

int psig(struct proc *p)
{
    int i;
    int stopsig[4] = {19, 20, 21, 22};
    int coresig[5] = {SIGQUIT, SIGABRT, SIGILL, SIGFPE, SIGSEGV};
    int termsig[7] = {SIGHUP, SIGINT, SIGPIPE, SIGALRM, SIGTERM, SIGUSR1, SIGUSR2};
    int termcalled = 0, corecalled = 0, stopcalled = 0;

    if (!p)
        return -1;

    acquire(&proc->lock);

    // 处理终止信号
    for (i = 0; i < 7; i++) {
        if (((p->sigpending & (1 << termsig[i])) == (1 << termsig[i])) &&
            ((p->sigblocked & (1 << termsig[i])) == 0) &&
            p->allinfo[termsig[i]].disposition == DFL) {
            if (termcalled == 0)
                term(p, &termcalled);
            p->sigpending &= ~(1 << termsig[i]);
        }
    }

    // 处理核心转储信号
    for (i = 0; i < 5; i++) {
        if (((p->sigpending & (1 << coresig[i])) == (1 << coresig[i])) &&
            ((p->sigblocked & (1 << coresig[i])) == 0) &&
            p->allinfo[coresig[i]].disposition == DFL) {
            if (corecalled == 0)
                core(p, &corecalled);
            p->sigpending &= ~(1 << coresig[i]);
        }
    }

    // 处理SIGSTOP信号
    if ((p->sigpending & (1 << 19)) == (1 << 19)) {
        stop(p, &stopcalled);
        p->sigpending &= ~(1 << 19);
    }

    // 处理SIGCHLD信号
    if (((p->sigpending & (1 << 17)) == (1 << 17)) &&
        ((p->sigblocked & (1 << 17)) == 0) &&
        p->allinfo[SIGCHLD].disposition == DFL) {
        ign();
        p->sigpending &= ~(1 << 17);
    }

    // 处理SIGCONT和停止信号
    for (i = 0; i < 4; i++) {
        if (((p->sigpending & (1 << stopsig[i])) == (1 << stopsig[i])) &&
            ((p->sigblocked & (1 << stopsig[i])) == 0) &&
            p->allinfo[stopsig[i]].disposition == DFL &&
            ((p->sigpending & (1 << 18)) == (1 << 18)) &&
            ((p->sigblocked & (1 << 18)) == 0) &&
            p->allinfo[SIGCONT].disposition == DFL) {
            p->sigpending &= ~(1 << 18);
            p->sigpending &= ~(1 << stopsig[i]);
        } else if (((p->sigpending & (1 << stopsig[i])) == (1 << stopsig[i])) &&
                   ((p->sigblocked & (1 << stopsig[i])) == 0) &&
                   p->allinfo[stopsig[i]].disposition == DFL) {
            if (stopcalled == 0)
                stop(p, &stopcalled);
            p->sigpending &= ~(1 << stopsig[i]);
        }
    }

    // 处理仅有的SIGCONT信号
    if (((p->sigpending & (1 << 18)) == (1 << 18)) &&
        ((p->sigblocked & (1 << 18)) == 0) &&
        p->allinfo[SIGCONT].disposition == DFL) {
        cont(p);
        p->sigpending &= ~(1 << 18);
    }

    // 处理用户定义的信号处理程序
    if (p->userdefed == 0) {
        release(&proc->lock);
        return 1;
    } else {
        for (i = 1; i < 32; i++) {
            if ((p->userdefed & (1 << i)) == (1 << i))
                break;
        }

        // 保存上下文到用户栈
        *(uint64 *)((p->trapframe->sp) - 8) = r_sstatus();
        *(uint64 *)((p->trapframe->sp) - 16) = p->trapframe->sp;
        *(uint64 *)((p->trapframe->sp) - 24) = r_scause();
        *(uint64 *)((p->trapframe->sp) - 32) = r_stval();
        *(uint64 *)((p->trapframe->sp) - 40) = r_sepc();
        *(uint64 *)((p->trapframe->sp) - 48) = p->trapframe->ra;
        *(uint64 *)((p->trapframe->sp) - 56) = p->trapframe->sp;
        *(uint64 *)((p->trapframe->sp) - 64) = p->trapframe->gp;
        *(uint64 *)((p->trapframe->sp) - 72) = p->trapframe->tp;
        *(uint64 *)((p->trapframe->sp) - 80) = p->trapframe->t0;
        *(uint64 *)((p->trapframe->sp) - 88) = p->trapframe->t1;
        *(uint64 *)((p->trapframe->sp) - 96) = p->trapframe->t2;
        *(uint64 *)((p->trapframe->sp) - 104) = p->trapframe->s0;
        *(uint64 *)((p->trapframe->sp) - 112) = p->trapframe->s1;
        *(uint64 *)((p->trapframe->sp) - 120) = p->trapframe->a0;
        *(uint64 *)((p->trapframe->sp) - 128) = p->trapframe->a1;
        *(uint64 *)((p->trapframe->sp) - 136) = p->trapframe->a2;
        *(uint64 *)((p->trapframe->sp) - 144) = p->trapframe->a3;
        *(uint64 *)((p->trapframe->sp) - 152) = p->trapframe->a4;
        *(uint64 *)((p->trapframe->sp) - 160) = p->trapframe->a5;
        *(uint64 *)((p->trapframe->sp) - 168) = p->trapframe->a6;
        *(uint64 *)((p->trapframe->sp) - 176) = p->trapframe->a7;
        *(uint64 *)((p->trapframe->sp) - 184) = p->trapframe->s2;
        *(uint64 *)((p->trapframe->sp) - 192) = p->trapframe->s3;
        *(uint64 *)((p->trapframe->sp) - 200) = p->trapframe->s4;
        *(uint64 *)((p->trapframe->sp) - 208) = p->trapframe->s5;
        *(uint64 *)((p->trapframe->sp) - 216) = p->trapframe->s6;
        *(uint64 *)((p->trapframe->sp) - 224) = p->trapframe->s7;
        *(uint64 *)((p->trapframe->sp) - 232) = p->trapframe->s8;
        *(uint64 *)((p->trapframe->sp) - 240) = p->trapframe->s9;
        *(uint64 *)((p->trapframe->sp) - 248) = p->trapframe->s10;
        *(uint64 *)((p->trapframe->sp) - 256) = p->trapframe->s11;
        *(uint64 *)((p->trapframe->sp) - 264) = p->trapframe->t3;
        *(uint64 *)((p->trapframe->sp) - 272) = p->trapframe->t4;
        *(uint64 *)((p->trapframe->sp) - 280) = p->trapframe->t5;
        *(uint64 *)((p->trapframe->sp) - 288) = p->trapframe->t6;

        *(uint64 *)((p->trapframe->sp) - 296) = i;
        *(uint64 *)((p->trapframe->sp) - 304) = (uint64)p->addr_sigreturn;
        p->trapframe->sp = p->trapframe->sp - 304;

        p->sigpending &= ~(1 << i);
        release(&proc->lock);

        __sync_synchronize();

        // 跳转到用户定义的信号处理程序
        jmptohandler(p->allinfo[i].handler, r_sstatus(), r_scause(), p->trapframe->sp, r_sstatus());
    }


    return -1;
}

void kernel_sigreturn(int signal)
{
    struct proc *p = myproc();
    if (!p)
        return;

    uint64 sstatus, scause, sepc, ra, sp, gp, tp, t0, t1, t2, s0, s1, a0, a1, a2, a3, a4, a5, a6, a7, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, t3, t4, t5, t6;
    sp = p->trapframe->sp; // 当前用户栈指针，指向信号处理程序参数

    t6 = *(uint64 *)(sp + 8);
    t5 = *(uint64 *)(sp + 16);
    t4 = *(uint64 *)(sp + 24);
    t3 = *(uint64 *)(sp + 32);
    s11 = *(uint64 *)(sp + 40);
    s10 = *(uint64 *)(sp + 48);
    s9 = *(uint64 *)(sp + 56);
    s8 = *(uint64 *)(sp + 64);
    s7 = *(uint64 *)(sp + 72);
    s6 = *(uint64 *)(sp + 80);
    s5 = *(uint64 *)(sp + 88);
    s4 = *(uint64 *)(sp + 96);
    s3 = *(uint64 *)(sp + 104);
    s2 = *(uint64 *)(sp + 112);
    a7 = *(uint64 *)(sp + 120);
    a6 = *(uint64 *)(sp + 128);
    a5 = *(uint64 *)(sp + 136);
    a4 = *(uint64 *)(sp + 144);
    a3 = *(uint64 *)(sp + 152);
    a2 = *(uint64 *)(sp + 160);
    a1 = *(uint64 *)(sp + 168);
    a0 = *(uint64 *)(sp + 176);
    s1 = *(uint64 *)(sp + 184);
    s0 = *(uint64 *)(sp + 192);
    t2 = *(uint64 *)(sp + 200);
    t1 = *(uint64 *)(sp + 208);
    t0 = *(uint64 *)(sp + 216);
    tp = *(uint64 *)(sp + 224);
    gp = *(uint64 *)(sp + 232);
    ra = *(uint64 *)(sp + 248);
    sepc = *(uint64 *)(sp + 256);
    scause = *(uint64 *)(sp + 264);
    sstatus = *(uint64 *)(sp + 280);
    sp = *(uint64 *)(sp + 272);

    restoreuser(t6, t5, t4, t3, s11, s10, s9, s8, s7, s6,
                s5, s4, s3, s2, a7, a6, a5, a4, a3, a2,
                a1, a0, s1, s0, t2, t1, t0, tp, gp, ra,
                sepc, scause, sstatus, sp);
}



// 指定信号的处理程序
// returnfn 指定了系统调用 sigreturn 的用户地址

sighandler_t
signal(addr_sigret returnfn, int signum, sighandler_t handler)
{
	if(signum <= 0 || signum > CURRNUM)
		return SIG_ERR;
		
	if(signum == SIGKILL || signum == SIGSTOP)
		return SIG_ERR;
	
	
	struct proc *p = myproc();
	
	if(!p)
		return SIG_ERR;
		
	sighandler_t prev;
	if(p->allinfo[signum].disposition == USERDEF)
		prev = p->allinfo[signum].handler;
	else if(p->allinfo[signum].disposition == IGN)
		prev = SIG_IGN;
	else
		prev = SIG_DFL;
	
	acquire(&proc->lock);

	
	p->addr_sigreturn = NULL;
	if(handler == SIG_IGN){
		if(p->allinfo[signum].disposition == USERDEF)
			p->userdefed = p->userdefed & (~(1 << signum));
		
		p->allinfo[signum].disposition = IGN;

	}
	else if(handler == SIG_DFL){
		if(p->allinfo[signum].disposition == USERDEF)
			p->userdefed = p->userdefed & (~(1 << signum));
			
		p->allinfo[signum].disposition = DFL;
		
	}
	else{
		p->allinfo[signum].disposition = USERDEF;
		p->userdefed = p->userdefed | (1 << signum);
		p->allinfo[signum].handler = handler;
		p->addr_sigreturn = returnfn;
	}
		
	release(&proc->lock);
	return prev;
}

//向进程或进程组发信号
int 
Kill(pid_t pid, int sig)
{

	if(sig <= 0 || sig > CURRNUM)
		return -1;
	
	if(pid == 1 || pid == 0 || pid < -1)
	{
		printf("Operation not allowed\n");
		return 0;
	}
	
	struct proc *p;
	struct proc *curproc = myproc();
	
	if(!curproc)
		return -1;
	
	int minusone = 0;
	
	if(pid == -1)
		minusone = 1;

	for(p = proc; p < &proc[NPROC]; p++){
    
		if(p->pid == 1)
    			continue;
    	
		if(minusone || p->pid == pid){
			acquire(&proc->lock);
			
			if(sig == SIGKILL){
	
				release(&proc->lock);
				kill(p->pid);
				
				if(minusone)
					continue;
				else
					return 0;
			}
			else if(sig == SIGCHLD && p->allinfo[SIGCHLD].disposition == DFL){
				release(&proc->lock);
				if(minusone)
					continue;
				else{
					return 0;
					
				}
			}
			else if(sig == SIGCONT){
				
				if(p->state == SLEEPING){
					p->state = RUNNABLE;
					p->justwoken = 1;
				}
				if(p->allinfo[SIGCONT].disposition != DFL && p->allinfo[SIGCONT].disposition != IGN){
					p->sigpending = p->sigpending | (1 << sig);
				}
				release(&proc->lock);
				if(minusone)
					continue;
				else
					return 0;
				
			}
	    	
			if(p->allinfo[sig].disposition != IGN && sig != 0){
				p->sigpending = p->sigpending | (1 << sig);
			}

			release(&proc->lock);
			if(minusone == 0)
				break;

		}
	}
	return 0;
}

int raise(int sig)
{
	struct proc *p = myproc();
	
	if(!p)
		return -1;
		
	return Kill(p->pid, sig);
	
}

//取得现有阻塞信号掩码
int
siggetmask(void)
{
	struct proc *p = myproc();
	if(!p)
		return -1;
	
	acquire(&proc->lock);
	uint mask = p->sigblocked;
	release(&proc->lock);
	
	return mask;
}

//用给定信号掩码替换现有阻塞信号掩码
int
sigsetmask(uint mask)
{
	

	if((mask & (1 << SIGKILL)) || (mask & (1 << SIGSTOP)) || (mask & (1 << SIGCONT)))
		return 0;
	if(mask & 1)
		return 0;
	
	struct proc *p = myproc();
	
	if(!p)
		return -1;
	
	
	acquire(&proc->lock);
	p->sigblocked = mask;
	release(&proc->lock);
	return mask;
}

//向被阻塞信号掩码中添加信号
int
sigblock(int sig)
{
	if(sig <= 0 || sig > CURRNUM)
		return -1;
		
	if(sig == SIGKILL || sig == SIGSTOP || sig == SIGCONT)
		return 0;
	uint mask = siggetmask();
	mask = mask | (1 << sig);
	return (sigsetmask(mask));
}


int
sigunblock(int sig)
{
	if(sig <= 0 || sig > CURRNUM)
		return -1;
		
	if(sig == SIGKILL || sig == SIGSTOP)
		return 0;
	uint mask = siggetmask();
	mask = mask & (~(1 << sig));
	return (sigsetmask(mask));
}
